"""Utility functions for distributions
"""


def check_is_integer(x):
    if not all(x == x.astype(int)):
        raise ValueError("All values of target must be integers!")


def check_is_positive(x):
    if not all(x >= 0):
        raise ValueError("All values of target must be positive!")


def check_is_gt_zero(x):
    if not all(x > 0):
        raise ValueError("All values of target must be > 0!")



"""conftest.py for xgboost_distribution.
"""
import pytest

import numpy as np
from sklearn.model_selection import train_test_split


@pytest.fixture
def small_X_y_data():
    """Small set of X, y data (single feature)"""

    def true_function(X):
        return np.sin(3 * X)

    def true_noise_scale(X):
        return np.abs(np.cos(X))

    np.random.seed(1234)
    n_samples = 100
    X = np.random.uniform(-2, 2, n_samples)
    y = true_function(X) + np.random.normal(scale=true_noise_scale(X), size=n_samples)

    return X[..., np.newaxis], y


@pytest.fixture
def small_train_test_data(small_X_y_data):
    """Small set of train-test split X, y data (single feature)"""
    X, y = small_X_y_data

    return train_test_split(X, y, test_size=0.2, random_state=1)



"""Negative binomial distribution
"""
import numpy as np
from scipy.special import digamma, expit
from scipy.stats import nbinom

from xgboost_distribution.distributions.base import BaseDistribution
from xgboost_distribution.distributions.utils import check_is_integer, check_is_positive


class NegativeBinomial(BaseDistribution):
    """Negative binomial distribution with log score

    Definition:

        f(k) = p^n (1 - p)^k binomial(n + k - 1, n - 1)

    with parameter (n, p), where n >= 0 and 1 >= p >= 0

    We reparameterize:

        n -> log(n) = a        |  e^a = n
        p -> log(p/(1-p)) = b  |  e^b = p / (1-p)   |  p = 1 / (1 + e^-b)

    The gradients are:

        d/da -log[f(k)] = -e^a [ digamma(k+e^a) - digamma(e^a) + log(p) ]
                        = -n   [ digamma(k+n) - digamma(n) + log(p) ]

        d/db -log[f(k)] = (k e^b - e^a) / (e^b + 1)
                        = (k - e^a e^-b) / (e^-b + 1)
                        = p * (k - e^a e^-b)
                        = p * (k - n e^-b)

    The Fisher Information:

        I(n) ~ p / [ n (p+1) ]
        I(p) = n / [ p (1-p)^2 ]

    where we used an approximation for I(n) presented here:
        http://erepository.uonbi.ac.ke:8080/xmlui/handle/123456789/33803

    In reparameterized form, we find I_r(n) and I_r(p):

        p / [ n (p+1) ] = I_r(n) [ d/dn log(n) ]^2
                        = I_r(n) ( 1/n )^2

        -> I_r(n) = np / (p+1)

        n / [ p (1-p)^2 ] = I_r(p) [ d/dp log(p/(1-p)) ]^2
                          = I_r(p) ( 1/ [ p (1-p) ] )^2

        -> I_r(p) = [ p^2 (1-p)^2 n ] / [ p (1-p)^2 ] = np

    Hence the reparameterized Fisher information:

        [  np / (p+1), 0 ]
        [  0,         np ]

    Ref:

        https://www.wolframalpha.com/input/?i=d%2Fda+-log%28+%5B1+%2F+%281+%2B+e%5E%28-b%29%29%5D+%5E%28e%5Ea%29+%281+-+%5B1+%2F+%281+%2B+e%5E%28-b%29%29%5D%29%5Ek+binomial%28%28e%5Ea%29+%2B+k+-+1%2C+%28e%5Ea%29+-+1%29+%29

    """

    @property
    def params(self):
        return ("n", "p")

    def check_target(self, y):
        check_is_integer(y)
        check_is_positive(y)

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        log_n, raw_p = params[:, 0], params[:, 1]
        n = np.exp(log_n)
        p = expit(raw_p)

        grad = np.zeros(shape=(len(y), 2))

        grad[:, 0] = -n * (digamma(y + n) - digamma(n) + np.log(p))
        grad[:, 1] = p * (y - n * (1 - p) / p)

        if natural_gradient:

            fisher_matrix = np.zeros(shape=(len(y), 2, 2))
            fisher_matrix[:, 0, 0] = (n * p) / (p + 1)
            fisher_matrix[:, 1, 1] = n * p

            grad = np.linalg.solve(fisher_matrix, grad)
            hess = np.ones(shape=(len(y), 2))  # we set the hessian constant

        else:
            raise NotImplementedError(
                "Normal gradients are currently not supported by this "
                "distribution. Please use natural gradients!"
            )

        return grad, hess

    def loss(self, y, params):
        n, p = self.predict(params)
        return "NegativeBinomialError", -nbinom.logpmf(y, n=n, p=p).mean()

    def predict(self, params):
        log_n, raw_p = params[:, 0], params[:, 1]
        n = np.exp(log_n)
        p = expit(raw_p)
        return self.Predictions(n=n, p=p)

    def starting_params(self, y):
        # TODO: starting params can matter a lot?
        return (np.log(np.mean(y)), 0)  # expit(0) = 0.5



import os

import pytest

import numpy as np
from sklearn.exceptions import NotFittedError

from xgboost_distribution.model import XGBDistribution


def test_XGBDistribution_early_stopping_fit(small_train_test_data):
    """Dummy test that everything runs"""

    X_train, X_test, y_train, y_test = small_train_test_data
    model = XGBDistribution(distribution="normal", max_depth=3, n_estimators=500)
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)], early_stopping_rounds=10)

    evals_result = model.evals_result()

    assert model.best_iteration == 6
    assert isinstance(evals_result, dict)


def test_XGBDistribution_early_stopping_predict(small_train_test_data):
    """Check that predict with early stopping uses correct ntrees"""
    X_train, X_test, y_train, y_test = small_train_test_data

    model = XGBDistribution(distribution="normal", max_depth=3, n_estimators=500)
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)], early_stopping_rounds=10)

    mean, var = model.predict(X_test)
    mean_iter, var_iter = model.predict(
        X_test, iteration_range=(0, model.best_iteration + 1)
    )
    np.testing.assert_array_equal(mean, mean_iter)
    np.testing.assert_array_equal(var, var_iter)


def test_distribution_set_param(small_X_y_data):
    """Check that updating the distribution params works"""

    X, y = small_X_y_data

    model = XGBDistribution(distribution="abnormal")
    model.set_params(**{"distribution": "normal"})

    model.fit(X, y)
    params = model.predict(X)

    assert isinstance(params[0], np.ndarray)


def test_XGBDistribution_save_and_load(small_X_y_data, tmpdir):
    X, y = small_X_y_data

    model = XGBDistribution(n_estimators=10)
    model.fit(X, y)
    preds = model.predict(X)

    model_path = os.path.join(tmpdir, "model.bst")
    model.save_model(model_path)

    saved_model = XGBDistribution()
    saved_model.load_model(model_path)
    saved_preds = saved_model.predict(X)

    np.testing.assert_array_equal(preds[0], saved_preds[0])
    np.testing.assert_array_equal(preds[1], saved_preds[1])


def test_predict_before_fit_fails(small_X_y_data):
    X, y = small_X_y_data
    model = XGBDistribution(n_estimators=10)

    with pytest.raises(NotFittedError):
        model.predict(X)



import re

from xgboost_distribution.distributions.base import BaseDistribution
from xgboost_distribution.distributions.log_normal import LogNormal  # noqa
from xgboost_distribution.distributions.negative_binomial import (  # noqa
    NegativeBinomial,
)
from xgboost_distribution.distributions.normal import Normal  # noqa
from xgboost_distribution.distributions.poisson import Poisson  # noqa

# TOTA: alternative way of importing distribution subclasses?
# __all__ = ["normal"]
# from . import *


def format_distribution_name(class_name):
    return re.sub(r"(?<!^)(?=[A-Z])", "-", class_name).lower()


AVAILABLE_DISTRIBUTIONS = {
    format_distribution_name(subclass.__name__): subclass
    for subclass in BaseDistribution.__subclasses__()
}


def get_distribution(name):
    """Get instantianted distribution based on name"""

    if name not in AVAILABLE_DISTRIBUTIONS.keys():
        raise ValueError(
            "Distribution is not implemented! Please choose one of "
            f"{set(AVAILABLE_DISTRIBUTIONS.keys())}"
        )

    return AVAILABLE_DISTRIBUTIONS[name]()


def get_distribution_doc():
    """Construct docstring for `distribution` param in XGBDistribution model"""

    param_doc = f"""
    distribution : {set(AVAILABLE_DISTRIBUTIONS.keys())}, default='normal'
        Which distribution to estimate. Available choices:
    """

    for name, subclass in AVAILABLE_DISTRIBUTIONS.items():
        param_doc += f"""
                - \"{name}\" - parameters: {subclass().params}\n"""

    param_doc += """
        Please see `scipy.stats` for a full description of the parameters of
        each distribution: https://docs.scipy.org/doc/scipy/reference/stats.html
    """

    return param_doc



"""Poisson distribution
"""
import numpy as np
from scipy.stats import poisson

from xgboost_distribution.distributions.base import BaseDistribution
from xgboost_distribution.distributions.utils import check_is_integer, check_is_positive


class Poisson(BaseDistribution):
    """Poisson distribution with log score

    Definition:

        f(k) = e^(-mu) mu^k / k!

    We reparameterize mu -> log(mu) = a to ensure mu >= 0. Gradient:

        d/da -log[f(k)] = e^a - k  = mu - k

    The Fisher information = 1 / mu, which needs to be expressed in the
    reparameterized form:

        1 / mu = I ( d/dmu log(mu) )^2 = I ( 1/ mu )^2

    Hence we find: I = mu

    """

    @property
    def params(self):
        return ("mu",)

    def check_target(self, y):
        check_is_integer(y)
        check_is_positive(y)

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        (mu,) = self.predict(params)

        grad = np.zeros(shape=(len(y), 1))
        grad[:, 0] = mu - y

        if natural_gradient:
            fisher_matrix = np.zeros(shape=(len(y), 1, 1))
            fisher_matrix[:, 0, 0] = mu

            grad = np.linalg.solve(fisher_matrix, grad)

            hess = np.ones(shape=(len(y), 1))  # we set the hessian constant
        else:
            hess = mu

        return grad, hess

    def loss(self, y, params):
        mu = self.predict(params)
        return "PoissonError", -poisson.logpmf(y, mu=mu).mean()

    def predict(self, params):
        log_mu = params
        mu = np.exp(log_mu)
        return self.Predictions(mu=mu)

    def starting_params(self, y):
        return (np.log(np.mean(y)),)



import pytest

import numpy as np

from xgboost_distribution.distributions import Normal


@pytest.fixture
def normal():
    return Normal()


@pytest.mark.parametrize(
    "y, params, natural_gradient, expected_grad",
    [
        (
            np.array([0, 0]),
            np.array([[0, 1], [1, 0]]),
            True,
            np.array([[0, 0.5], [1, 0]]),
        ),
        (
            np.array([0, 0]),
            np.array([[0, 1], [1, 0]]),
            False,
            np.array([[0, 1], [1, 0]]),
        ),
    ],
)
def test_gradient_calculation(normal, y, params, natural_gradient, expected_grad):
    grad, hess = normal.gradient_and_hessian(
        y, params, natural_gradient=natural_gradient
    )
    np.testing.assert_array_equal(grad, expected_grad)



"""Distribution base class
"""
from abc import ABC, abstractmethod
from collections import namedtuple


class BaseDistribution(ABC):
    """Base class distribution for XGBDistribution.

    Note that by design all distributions are **stateless**.
    A distribution is thus a collection of functions that operate on the data
    (`y`) and the outputs of the xgboost Booster (`params`).
    """

    def __init__(self):
        self.Predictions = namedtuple("Predictions", (p for p in self.params))

    def check_target(self, y):
        pass

    @property
    @abstractmethod
    def params(self):
        pass

    @abstractmethod
    def starting_params(self, y):
        pass

    @abstractmethod
    def gradient_and_hessian(self, y, params):
        pass

    @abstractmethod
    def loss(self, y, params):
        pass

    @abstractmethod
    def predict(self, params):
        pass
